"""Mypy stubs for pdcast/util/round/tolerance.pyx"""
from decimal import Decimal

from pdcast.util.type_hints import numeric


class Tolerance:
    real: Decimal
    imag: Decimal

    def __init__(self, tol: numeric) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __bool__(self) -> bool: ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __complex__(self) -> complex: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
