
/* admonition toggle button styling */
button.toggle-button {
    color: white;
}


/* container toggle button styling */
.toggle-button summary {
    color: white;
}


/* toggle text positioning */
.toggle-button::before {
    vertical-align: top;
    position: relative;
    top: 4px;
}
